from setuptools import setup, find_packages
setup(name='txgossip',
      version='0.1',
      description='simple cluster communication protocol',
      author='Johan Rydberg',
      author_email='johan.rydberg@gmail.com',
      url='http://github.com/jrydberg/txgossip',
      packages=find_packages()
)
